import subprocess
from websockets.client import connect, WebSocketClientProtocol
import asyncio
import json
import time
from collections import defaultdict, namedtuple
from typing import Callable, Any, Optional, TypeAlias, ClassVar, Union, Iterable
from enum import Enum
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from datetime import datetime, timedelta


class MsgEnv(Enum):
    private = 0
    group = 1

class AtTarget(Enum):
    all = 0
    sender = 1


class Controller():
    def __init__(self, bot: 'Bot', data: dict[str, Any]) -> None:
        self._bot = bot
        self._data = data
        self._msgEnv = None

        self._target_id = None

        self._cmd = None

        self._send_list: list[dict[str, Any]] = []

    def setEnv(self, env: MsgEnv):
        self._msgEnv = env

    def setTarget(self, target: int):
        self._target_id = target

    def setCmd(self, cmd: str):
        self._cmd = cmd
    
    def text(self, text: str):
        self._send_list.append({
            'type': 'text',
            'data': {
                'text': text
            }
        })
        return self
    
    def at(self, id):
        self._send_list.append({
            'type': 'at',
            'data': {
                'qq': id
            }
        })
        return self
    
    def atAll(self):
        self._send_list.append({
            'type': 'at',
            'data': {
                'qq': 'all'
            }
        })
        return self
    
    def atSender(self):
        self._send_list.append({
            'type': 'at',
            'data': {
                'qq': self.sender_id()
            }
        })
        return self
    
    def reply(self):
        self._send_list.append({
            'type': 'reply',
            'data': {
                'id': self.msg_id()
            }
        })
        return self
    
    def sender_id(self) -> int:
        return self._data['user_id']
    
    def sender_name(self) -> str:
        return self._data['sender']['nickname']

    def msg_id(self) -> int:
        return self._data['message_id']
    
    def msg(self) -> str:
        if self._cmd is None:
            return self._data['raw_message']
        else:
            return self._data['raw_message'].split(maxsplit=1)[1]

    def send(self):
        match self._msgEnv:
            case MsgEnv.private:
                self._sendPrivate()
            case MsgEnv.group:
                self._sendGroup()

    def _beforeSendPrivate(self):
        pass

    def _sendPrivate(self):
        self._beforeSendPrivate()
        if self._target_id is None:
            try:
                self._target_id = self._data['user_id']
            except KeyError:
                raise Exception('wrong msg environment')
                
        action = 'send_private_msg'
        params = {
            'user_id': self._target_id,
            'message': self._send_list
        }
        self._bot.addToSend((action, params))
    
    def _beforeSendGroup(self):
        pass

    def _sendGroup(self):
        self._beforeSendGroup()
        if self._target_id is None:
            try:
                self._target_id = self._data['group_id']
            except KeyError:
                raise Exception('wrong msg environment')
            
        action = 'send_group_msg'
        params = {
            'group_id': self._target_id,
            'message': self._send_list
        }
        self._bot.addToSend((action, params))


    
    
class Option(ABC):
    @abstractmethod
    def meet(self, data: dict[str, Any], ctl: Controller) -> bool:
        pass

@dataclass(repr=False, eq=False)
class Env(Option):
    _p: bool = False
    _g: bool = False
    
    def meet(self, data, ctl) -> bool:
        if data['message_type'] == 'private' and self._p:
            ctl.setEnv(MsgEnv.private)
            return True
        if data['message_type'] == 'group' and self._g:
            ctl.setEnv(MsgEnv.group)
            return True
        return False

    def allowPrivate(self):
        self._p = True

    def allowGroup(self):
        self._g = True

@dataclass(repr=False, eq=False)
class Private(Option):
    id: list[int] | None = None
    reply: bool = False

    def meet(self, data, ctl) -> bool:
        if data['message_type'] != 'private':
            return False
        if self.id is not None and data['user_id'] not in self.id:
            return False
        if self.reply:
            ctl.reply()
        return True

@dataclass(repr=False, eq=False)
class Group(Option):
    id: list[int] | None = None
    at: AtTarget | int | None = None
    reply: bool = True

    def meet(self, data, ctl) -> bool:
        if data['message_type'] != 'group':
            return False
        if self.id is not None and data['group_id'] not in self.id:
            return False
        if self.reply:
            ctl.reply()
        match self.at:
            case AtTarget.all:
                ctl.atAll()
            case AtTarget.sender:
                ctl.atSender()
            case None:
                pass
            case id:
                ctl.at(id)
            
        return True

@dataclass(repr=False, eq=False)
class Command(Option):
    cmd: str
    
    def meet(self, data, ctl) -> bool:
        raw_msg: str = data['raw_message']
        ctl.setCmd(self.cmd)
        return raw_msg.split(maxsplit=1)[0].startswith(self.cmd)

@dataclass(repr=False, eq=False)
class Keyword(Option):
    word: str
    
    def meet(self, data, ctl):
        return self.word in data['raw_message']

@dataclass(repr=False, eq=False)
class ScheduledTask():
    func: Callable[[Controller], None]
    options: list[Option]
    date: datetime
    cycle: timedelta | None
    count: int
    
    def refresh(self):
        if self.cycle and self.count > 1:
            self.date += self.cycle
            self.count -= 1
            return self.date
        else:
            return None
        
    def run(self, ctl: Controller):
        for option in self.options:
            match option:
                case Private(ids, _):
                    if ids is None:
                        continue
                    ctl.setEnv(MsgEnv.private)
                    for id in ids:
                        ctl.setTarget(id)
                        self.func(ctl)
                case Group(ids, at, _):
                    if ids is None:
                        continue
                    ctl.setEnv(MsgEnv.group)
                    for id in ids:
                        ctl.setTarget(id)
                        if at == AtTarget.all:
                            ctl.atAll()
                        self.func(ctl)
                case _:
                    pass

        

class Bot():
    def __init__(self, websocket_port: int) -> None:

        self._websockets_port = websocket_port

        self._detects: dict[Callable[[Controller], None], list[Option]] = {}
        self._schedules: asyncio.PriorityQueue[tuple[datetime, ScheduledTask]] = asyncio.PriorityQueue()

        self._to_send: asyncio.Queue[tuple[str, dict[str, Any]]] = asyncio.Queue()

    def run(self, stop_on_kb_interrupt=True):
            print(self._detects)
            asyncio.run(self._run(stop_on_kb_interrupt))

    async def _run(self, stop_on_kb_interrupt):
        async for ws in connect(f'ws://127.0.0.1:{self._websockets_port}', open_timeout=1):
            try:
                await asyncio.gather(self._recive(ws), self._send(ws), self._checkSchedules())
            except KeyboardInterrupt:
                if stop_on_kb_interrupt:
                    break
            except Exception as e:
                print('exception: ', e)
                continue

    
    async def _recive(self, ws: WebSocketClientProtocol):
        while True:
            data: dict[str, Any] = json.loads(await ws.recv())
            match data.get('post_type'):
                case 'message':
                    self.handleMessage(data)
                case _:
                    pass

    async def _send(self, ws: WebSocketClientProtocol):
        while True:
            try:
                action, params = await self._to_send.get()
                print(params)
                await ws.send(json.dumps({
                    'action': action,
                    'params': params
                }))
            except Exception as e:
                print(e)
    
    async def _checkSchedules(self):
        while True:
            schedule_time, task = await self._schedules.get()
            if 0 < (time.time()-schedule_time.timestamp()) < 0.1:
                task.run(Controller(self, {}))
                if new_date := task.refresh():
                    await self._schedules.put((new_date, task))
            else:
                await self._schedules.put((schedule_time, task))
                await asyncio.sleep(0.01)
        
    def handleMessage(self, data: dict[str, Any]):
        print(data)
        for func, Options in self._detects.items():
            ctl = Controller(self, data)
            meet = True
            for Option in Options:
                if not Option.meet(data, ctl):
                    meet = False
                    break
            if meet:
                func(ctl)

    def addToSend(self, data: tuple[str, dict[str, Any]]):
        print(data)
        self._to_send.put_nowait(data)
    
    def detect(self, options: list[Option]):
        def f(func: Callable[[Controller], None]):
            env_option = Env()
            for option in options:
                match option:
                    case Private():
                        env_option.allowPrivate()
                    case Group():
                        env_option.allowGroup()
            options.append(env_option)
            self._detects[func] = options
            return func
        return f
    
    def schedule(self, options: list[Option], date: datetime, cycle: timedelta | None = None, count: int = 1):
        def f(func: Callable[[Controller], None]):
            self._schedules.put_nowait((date, ScheduledTask(func, options, date, cycle, count)))
            return func
        return f
    
    

if __name__ == '__main__':
    bot = Bot(3001)

    @bot.detect([
        Private(),
        Command('#echo')
    ])
    def echo(ctl: Controller):
        ctl.text(ctl.msg()).send()

#     @bot.detect([
#         Private([2155891548, 192847693])
#     ])
#     def ssb(ctl: Controller):
#         ctl.text('什么玩意').send()

#     @bot.detect([
#         Private([2155891548])
#     ])
#     def sb(ctl: Controller):
#         ctl.text('傻逼').send()

    

    """ @bot.detect([
        Group(),
        Command('say')
    ])
    def say(ctl: Controller):
        ctl.text('hello~').text(ctl.sender_name()).send() """

    """ @bot.schedule([
        Private([2155891548]),
        Group([906519909])
    ], datetime.now()+timedelta(seconds=3), timedelta(seconds=3), 5)
    def test(ctl: Controller):
        ctl.text('aaaaaaaa').send() """
    
    bot.run()